; SketchyLISP Example Program
; Copyright (C) 2005,2006,2007 Nils M Holm. All rights reserved.
; See the file LICENSE of the SketchyLISP distribution
; for conditions of use.

; ---name---
; evaluate

; ---purpose---
; Evaluate arithmetic expressions.
; Evaluate all operations with numeric arguments
; (or subexpressions evaluating to numeric arguments)
; to their results. Reduce expressions with one variable
; argument using rules such as
; .B
; .C X+0 -> X
; .B
; and
; .B
; .C X^0 -> 1.
; .B
; Return the evaluated expression.
; .P
; The following functions are recognized:
; .B
; .C + - * quotient expt
; .P
; Variables are represented by single letters.

; ---language---
; R5RS Scheme

; ---keywords---
; EVALUATE function, evaluation, partial, math expressions
; arithmetic expressions, arithmetics

; ---args---
; X - expression

; ---example---
; (evaluate '(* x (expt 2 (+ 3 4)))) => (* x 128)

; (require "lib/list.scm")
; (require "lib/caar.scm") ; cadr
; (require "lib/caaar.scm") ; caddr
; (require "lib/zerop.scm") ; zero?
; (require "lib/plus.scm") ; +
; (require "lib/minus.scm") ; -
; (require "lib/times.scm") ; *
; (require "lib/quotient.scm")
; (require "lib/expt.scm")
; (require "lib/member.scm")

; ---code---
(define (evaluate x)
  (letrec
    ; Evaluate (fn left-num right-num)
    ((compute
       (lambda (op left right)
         (cond ((eq? op '+) (+ left right))
               ((eq? op '-) (- left right))
               ((eq? op '*) (* left right))
               ((eq? op 'quotient) (quotient left right))
               ((eq? op 'expt) (expt left right))
               (else (bottom 'compute)))))
     (zero-p
       (lambda (x)
         (and (not (symbol? x))
              (zero? x))))
     (one-p
       (lambda (x)
         (and (not (symbol? x))
              (= 1 x))))
     (_evaluate
       (lambda (x)
         (cond
           ; Pass thru symbols and numbers
           ((symbol? x) x)
           ((number? x) x)
           ; Evaluate or reduce binary functions
           ((member (car x) '(+ - * quotient expt))
             ; First evaluate the arguments
             (let ((left (_evaluate (cadr x)))
                  (right (_evaluate (caddr x))))
               (cond
                 ; Both arguments constant,
                 ; compute the result
                 ((and (number? left) (number? right))
                   (compute (car x) left right))
                 ((eq? (car x) '+)
                   (cond ((zero-p right) left)   ; X+0 => X
                     ((zero-p left) right)       ; 0+X => X
                     (else (list (car x) left right))))
                 ((eq? (car x) '-)
                   (cond ((zero-p right) left)   ; X-0 => X
                     (else (list (car x) left right))))
                 ((eq? (car x) '*)
                   (cond ((zero-p right) 0)    ; X*0 => 0
                     ((zero-p left) 0)         ; 0*X => 0
                     ((one-p left) right)      ; X*1 => X
                     ((one-p right) left)      ; 1*X => X
                     (else (list (car x) left right))))
                 ((eq? (car x) 'quotient)
                   (cond ((zero-p right)     ; X/0 => _|_
                       (bottom 'divide-by-zero))
                     ((zero-p left) 0)       ; 0/X => 0
                     ((one-p right) left)    ; X/1 => X
                     (else (list (car x) left right))))
                 ((eq? (car x) 'expt)
                   (cond ((zero-p right) 1)    ; X^0 => 1
                     ((zero-p left) 0)         ; 0^X => 0
                     ((one-p left) 1)          ; 1^X => 1
                     ((one-p right) left)      ; X^1 => X
                     (else (list (car x) left right))))
                 ; Unknown operator?
                 (else (bottom 'this-cannot-happen)))))
           (else (bottom (list 'unknown-operator: x)))))))
    (_evaluate x)))
